/*
 * Share.java
 *
 * Created on October 19, 2002, 11:21 AM
 */
package Secret_Share;

import java.math.BigInteger;
import java.io.*;

/** <P>This class stores the individual shares of the secret. Primarily it stores the share number, which is also the x value used in the generating functions. The share value is equal to f(x). There is also a commitment value equal to G(x)</P>
 * <P>They are all individually verifiable via a public version. Each share stores a copy of the public version of all n of the shares. The shares can also be verified via Pedersen's "E Array".</P>
 * <P>Version .2 added a copy function. It is needed for share recovery.</P>
 * <P>As of version .3, the share class functions correctly for share recovery.</P>
 * <P>Version .4 added getSerialized().</P>
 * @author Brice Daniels
 * @version .4
 */
public class Share implements Serializable {
    
    private BigInteger p = null;
    
    private BigInteger q = null;
    
    private BigInteger g = null;
    
    private BigInteger h = null;
    
    private int shareNumber = 0;
    
    private BigInteger secretShare = null;
    
    private BigInteger secretCommitment = null;
    
    private BigInteger[] publicShareVersions = null;
    
    private int k = 0;
    
    /** <P>Creates a new instance of <CODE>Share</CODE>.</P>
     * <P>The prime modulus p must have the property that p-1 is divisible by q. Usually p-1 = (2 or 4) * q. P is used for the verifibility scheme while q is used as the sharing modulus.</P>
     * <P>The group members g and h must be in the sub-group of p generated by q. This is verified by raising g (or h) to the qth power and verifying that it is congruent to 1 mod p. G and h are used in the verifiability scheme.</P>
     * @param k The minimum number of required shares to recover the secret.
     * @param x The share number
     * @param s This share's portion of the secret
     * @param t The confirmation value for this secret share
     * @param p The prime modulus p.
     * @param q The prime modulus q
     * @param g The group member g.
     * @param h The group member h.
     * @param y The public version of all of the secret shares
     * @throws ShareException Indication of an error during share creation. Specific information may be available in the exception's message.
     */
    public Share(int x, BigInteger s, BigInteger t, BigInteger p, BigInteger q, BigInteger g, BigInteger h, BigInteger[] y, int k)
    throws ShareException {
        if(x <= 0) {
            throw new ShareException("Share value cannot be less than or equal to 0");
        }
        if(s == null) {
            throw new ShareException("Secret Share is null");
        }
        if(t == null) {
            throw new ShareException("Confirmation Share is null");
        }
        if(p == null) {
            throw new ShareException("P is null");
        }
        if(q == null) {
            throw new ShareException("Q is null");
        }
        if(g == null) {
            throw new ShareException("G is null");
        }
        if(h == null) {
            throw new ShareException("H is null");
        }
        if(y == null) {
            throw new ShareException("Y is null");
        }
        this.shareNumber = x;
        this.secretShare = s;
        this.secretCommitment = t;
        this.p = p;
        this.q = q;
        this.g = g;
        this.h = h;
        this.k = k;
        this.publicShareVersions = y;
    }
    
    /** <P>Returns the share number of the current share. The share number is equal to the value used to generate the share in the original polynomial, eg share number 3 would mean that f(3) and g(3) were used to generate this share.</P>
     * @return The share's share number.
     */
    public int getShareNumber() {
        return this.shareNumber;
    }
    
    /** <P>The secret share, equal to f(x).</P>
     * @return The secret share as a <CODE>BigInteger</CODE>.
     */
    public BigInteger getSecretShare() {
        return this.secretShare.add(BigInteger.ZERO);
    }
    
    /** <P>The secret commitment, equal to g(x).</P>
     * @return The secret commitment value as a <CODE>BigInteger</CODE>.
     */
    public BigInteger getSecretCommitment() {
        return this.secretCommitment.add(BigInteger.ZERO);
    }
    
    /** <P>Returns the public version of this share.</P>
     * @return A <CODE>BigInteger</CODE> equal to g<SUP>s</SUP>*h<SUP>t</SUP> (mod p).
     */
    public BigInteger getPublicVersion() {
        return this.publicShareVersions[this.shareNumber - 1];
    }
    
    /** <P>Return an array of length n with all of the public versions for all n share values.</P>
     * @return A <CODE>BigInteger</CODE> array with all of the public versions.
     */    
    public BigInteger[] getAllPublicVersion() {
        BigInteger[] returnPublicShares = new BigInteger[this.publicShareVersions.length];
        for(int i = 0;i < this.publicShareVersions.length;i++) {
            returnPublicShares[i] = this.publicShareVersions[i].add(BigInteger.ZERO);
        }
        return returnPublicShares;
    }
    
    /** <P>Return the g value used by the current share. This function does not return a reference to the value contained by share, just a copy of it.</P>
     * @return The g value.
     */    
    public BigInteger getG() {
        return this.g.add(BigInteger.ZERO);
    }
    
    /** <P>Return the h value used by the current share. This function does not return a reference to the value contained by share, just a copy of it.</P>
     * @return The h value.
     */    
    public BigInteger getH() {
        return this.h.add(BigInteger.ZERO);
    }
    
    /** <P>The p value used by the current share. This function does not return a reference to the value contained by share, just a copy of it.</P>
     * @return The p value.
     */    
    public BigInteger getP() {
        return this.p.add(BigInteger.ZERO);
    }
    
    /** <P>The q value used by the current share. This function does not return a reference to the value contained by share, just a copy of it.</P>
     * @return The q value.
     */    
    public BigInteger getQ() {
        return this.q.add(BigInteger.ZERO);
    }
    
    /** <P>The minimum number of shares required to recover the original secret. This is the same value that was entered when the secret was originally shared.</P>
     * @return The value k.
     */
    public int getK() {
        return this.k;
    }
    
    /** <P>Returns the number of shares according to this share. It is based on the <CODE>publicShareVersion.length</CODE>.</P>
     * @return The number of shares. */    
    public int getN() {
        return this.publicShareVersions.length;
    }
    
    /**
     * @param out
     * @throws IOException  */
    private void writeObject(java.io.ObjectOutputStream out) throws IOException {
        out.writeInt(this.shareNumber);
        out.writeObject(this.secretShare);
        out.writeObject(this.secretCommitment);
        out.writeObject(this.publicShareVersions);
        out.writeObject(this.p);
        out.writeObject(this.q);
        out.writeObject(this.g);
        out.writeObject(this.h);
        out.writeInt(this.k);
    }
    
    /** <P>This method serializes the current object and returns that object in the form of a byte array. This is often very useful for encryption.</P>
     * @throws IOException Indicates that an I/O error occured while trying to serialize this class.
     * @return A <CODE>byte</CODE> array containing the serialized version of this object.
     */
    public byte[] getSerialized() throws IOException {
        ByteArrayOutputStream bStream = new ByteArrayOutputStream();
        ObjectOutputStream ObjStream = new ObjectOutputStream(bStream);
        
        ObjStream.writeObject(this);
        ObjStream.close();
        return bStream.toByteArray();
    }
    
    /**
     * @param in
     * @throws IOException
     * @throws ClassNotFoundException  */
    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        this.shareNumber = in.readInt();
        this.secretShare = (BigInteger) in.readObject();
        this.secretCommitment = (BigInteger) in.readObject();
        this.publicShareVersions = (BigInteger[]) in.readObject();
        this.p = (BigInteger) in.readObject();
        this.q = (BigInteger) in.readObject();
        this.g = (BigInteger) in.readObject();
        this.h = (BigInteger) in.readObject();
        this.k = in.readInt();
    }
    
    /** <P>This method provides a way to add a subshare to a current share. The two shares must have the same basis (global values of p, q, g and h) and share number. The addition adds the value of the <CODE>addedShare.getSecretShare()</CODE> to the current secret share and <CODE>addedShare.getSecretCommitment()</CODE> to the secret commitment share.</P>
     * <P>Since adding two shares would cause the current share to be invalid against the public versions held by this share and all of the other shares from the splitting. Therefore, the commitment values that were generated along with the sub share are added into the public version held by this share for all shares. It is therefore necessary to make sure that all shares add the same set of sub-shares so that they all end up with the same public versions in the end.</P>
     * @param commitmentValue A <CODE>BigInteger</CODE> array of length <CODE>k</CODE> that was created by the split function used to create the subshare.
     * @param addedShare The specific share to add to this share
     * @throws ShareException Indicates invalid commitment value length in the <CODE>addedShare</CODE>
     * @throws ShareUtilException Indicates an error occurred with the share utilities called by this method.
     * @throws ShareAddException Indicates an error with the arguments to the add function.
     */
    public void add(Share addedShare, BigInteger[] commitmentValue) 
    throws ShareAddException, ShareUtilException, ShareException {
        if(this.shareNumber != addedShare.shareNumber) {
            throw new ShareAddException("Incorrect Share Number");
        }
        if(!this.relativeBasis(addedShare)) {
            throw new ShareAddException("added share is not relatively equal");
        }
        if(!addedShare.verify(commitmentValue)) {
            throw new ShareAddException("added share failed self-integrity check");
        }
        
        BigInteger newSecretShare = null;
        BigInteger newSecretCommitment = null;
        BigInteger[] newPublicShareVersions = null;
        
        try {
            //update the secret share
            newSecretShare = this.secretShare.add(addedShare.getSecretShare());
            newSecretShare = newSecretShare.mod(this.q);
            
            //update the secret share commitment
            newSecretCommitment = this.secretCommitment.add(addedShare.getSecretCommitment());
            newSecretCommitment = newSecretCommitment.mod(this.q);
            
            //updates the public shares stored by this share
            newPublicShareVersions = new BigInteger[this.publicShareVersions.length];
            BigInteger commitmentValueExponent;
            BigInteger commitmentValuePower;
            for(int i = 0;i < newPublicShareVersions.length;i++) {
                newPublicShareVersions[i] = new BigInteger(this.publicShareVersions[i].toByteArray());
                for(int m = 0;m < commitmentValue.length;m++) {
                    commitmentValueExponent = new BigInteger(Integer.toString((int)Math.pow((double)(i+1),(double)m)));
                    commitmentValuePower = commitmentValue[m].modPow(commitmentValueExponent, this.p);
                    newPublicShareVersions[i] = newPublicShareVersions[i].multiply(commitmentValuePower);
                    newPublicShareVersions[i] = newPublicShareVersions[i].mod(this.p);
                }
            }
        } catch (ArithmeticException e) {
            throw new ShareUtilException("Arithmetic error during update");
        }
        
        this.secretShare = newSecretShare;
        this.secretCommitment = newSecretCommitment;
        this.publicShareVersions = newPublicShareVersions;
    }
    
    /** <P>Compares this share with another one. This is necessary to make sure that two shares came from the same splitting or one of them was part of share recovery or share renewal.</P>
     * <P>The two shares have a relative basis if they have the same p, q, g and h values.</P>
     * @param comparedShare The share you are comparing.
     * @return True if and only if the current share and <CODE>compareShare</CODE> have the same public values.
     */
    public boolean relativeBasis(Share comparedShare) {
        boolean returnValue = true;
        
        if(!this.g.equals(comparedShare.getG())) {
            returnValue = false;
        }
        if(!this.h.equals(comparedShare.getH())) {
            returnValue = false;
        }
        if(!this.p.equals(comparedShare.getP())) {
            returnValue = false;
        }
        if(!this.q.equals(comparedShare.getQ())) {
            returnValue = false;
        }
        
        return returnValue;
    }
    
    /** <P>This function compares the current value contained by the share against the <CODE>publicShareVersion</CODE> contained by the current share.</P>
     * @return true if the values stored in this share are equivalent to the public version contained within this share.
     * @throws ShareUtilException Indicates an error in calculating the public version.
     */
    public boolean selfVerify() throws ShareUtilException {
        BigInteger calculatedPublicShare = ShareUtil.calculatePublicVersion(this.secretShare, this.secretCommitment, this.p, this.g, this.h);
        return this.publicShareVersions[this.shareNumber - 1].equals(calculatedPublicShare);
    }
    
    /** <P>Verifies the current share based on the commitment values that are passed as arguments to the split function. The key use of this function is to verify that the "dealer" of the shares sent shares generated by the functions f and g.</P>
     * <P>The commitment values should be the E array used by Pedersen in his VSS protocol. Essentially, value i in the array should be: g^a[i]*h^b[i] mod p where a[*] and b[*] are the coefficients of f(x) and g(x).</P>
     * @return True if the share is consistent with the submitted E array.
     * @param commitmentValue Pedersen's E array
     * @throws ShareUtilException Indicates an error in verifying the share. 
     * @throws ShareException Indicates that the <CODE>commitmentValue</CODE> array is the wrong length
     */
    public boolean verify(BigInteger[] commitmentValue) throws ShareUtilException, ShareException {
        if(commitmentValue.length != this.k) {
            throw new ShareException("invalid commitmentValue length");
        }
        boolean verified = false;
        if(this.selfVerify()) {
            BigInteger commitmentValueExponent;
            BigInteger commitmentValuePower;
            BigInteger calculatedPublicShare = BigInteger.ONE;
            for(int j = 0;j < commitmentValue.length;j++) {
                commitmentValueExponent = new BigInteger(Integer.toString((int)Math.pow((double)this.shareNumber,(double)j)));
                commitmentValuePower = commitmentValue[j].modPow(commitmentValueExponent, this.p);
                calculatedPublicShare = calculatedPublicShare.multiply(commitmentValuePower);
                calculatedPublicShare = calculatedPublicShare.mod(this.p);
            }
            
            verified = ShareUtil.calculatePublicVersion(this.secretShare,this.secretCommitment,this.p,this.g,this.h).equals(calculatedPublicShare);
        }
        
        return verified;
    }
    
    /** <P>This function verifies a questionable share against the public version stored locally. This function is primarily used by <CODE>ShareUtil.recover</CODE>.</P>
     * @return true if the questioned
     * @param questionedShare The share you are trying to verify against the public version held within this share.
     * @throws ShareException Indication of an error with the argument.
     * @throws ShareUtilException Indication of an error attempting to calculate the public version of the <CODE>questionedShare</CODE>.
     */    
    public boolean verifyShare(Share questionedShare) throws ShareException, ShareUtilException {
        if(questionedShare == null) {
            throw new ShareException("questionedShare is null");
        }
        BigInteger calculatedPublicShare = ShareUtil.calculatePublicVersion(questionedShare.getSecretShare(), questionedShare.getSecretCommitment(), this.p, this.g, this.h);
        return this.publicShareVersions[questionedShare.getShareNumber() - 1].equals(calculatedPublicShare);
    }
    
    /** <P>Creates an exact copy of the share.</P>
     * @throws ShareException Indication of an error during share creation. Specific information may be available in the exception's message.
     * @return A copy of the <CODE>Share</CODE>.
     */    
    public Share copy() throws ShareException {
        return new Share(this.shareNumber,this.secretShare,this.secretCommitment,this.p,this.q,this.g,this.h,this.publicShareVersions,this.k);
    }
    
    /** <P>This is not actually a normal function for the share, it creates random values for the secret share and the commitment share.</P>
     * @param numBits The number of bits for the secret share and the secret commitment.
     */    
    public void corruptShare(int numBits) {
        java.security.SecureRandom sr = new java.security.SecureRandom();
        this.secretShare = new BigInteger(numBits,sr);
        this.secretShare = this.secretShare.mod(this.q);
        this.secretCommitment = new BigInteger(numBits,sr);
        this.secretCommitment = this.secretCommitment.mod(this.q);
    }
}
