/*
 * ShareUtil.java
 *
 * Created on October 19, 2002, 12:35 PM
 */

package Secret_Share;

import java.math.BigInteger;
import java.security.SecureRandom;

/** <P>The Share Utilities contain the necessary auxilliary functions of splitting and recovery. These functions can be used in accordance with Jarecki's PSS protocols to achieve proactive secret sharing.</P>
 * <P>The function for adding two shares works with share renewal (adding shares of zero). Share renewal works in versions .2 and higher.</P>
 * <P>These share functions can be used for share recovery. As of version .3, share recovery works in a (5,4) scheme at 128 bits.</P>
 * @author Brice Daniels
 * @version .5
 */
public class ShareUtil {
    
    /** <P>This is a simplification method used to hide the last argument, r, of the other split method. All that this method does is call the other split method with r set to null (and therefore ignored).</P>
     * <P>This method can be used for all normal instances of secret splitting. As it calls the other method, this method has the same properties with respect to the secret being zero (as the first coefficient of g(x) will also be set to zero).</P>
     * @return The array of shares
     * @param commitmentValues This should be an empty <CODE>BigInteger</CODE> array of length <CODE>k</CODE>.
     * @param secret The <CODE>BigInteger</CODE> representation of the secret.
     * @param n The number of shares to generate.
     * @param k The threshold of shares required to recover the secret.
     * @param numBits The minimum number of bits to use in the randomly generated polynomials (actual bit sizes for shares may be larger).
     * @param p The modulus p. It must be a prime number such that q divides (p-1). P is used in verification of shares.
     * @param q The modulus q. It must be a prime number fitting the requirement for p. This is the primary prime used in the sharing of the secret.
     * @param g G is a member of the subgroup generated by q in the field Z[p]. G is used in the verification of shares.
     * @param h H must also be in the same subgroup as g. It is used in verification of shares.
     * @throws ShareUtilException Indicates an error with the various other share utilities called by the <CODE>split</CODE> function.
     * @throws SplitException Indicates an error in the split process. Additional information is available from the exception message.
     */
    public static Share[] split(BigInteger secret, int n, int k, int numBits, BigInteger p, BigInteger q, BigInteger g, BigInteger h, BigInteger[] commitmentValues)
    throws SplitException, ShareUtilException {
        return ShareUtil.split(secret,n,k,numBits,p,q,g,h,commitmentValues,null);
    }
    
    /** <P>This is the primary function used to split a secret. It is called by the other split method.</P>
     * <P>If the secret is zero, then the first coefficient of G(x) will be set to zero. This is required for Share Renewal.</P>
     * @return The array of shares.
     * @param commitmentValues This should be an empty <CODE>BigInteger</CODE> array of length <CODE>k</CODE>.
     * @param secret The <CODE>BigInteger</CODE> representation of the secret.
     * @param n The number of shares to generate.
     * @param k The threshold of shares required to recover the secret.
     * @param numBits The minimum number of bits to use in the randomly generated polynomials (actual bit sizes for shares may be larger).
     * @param p The modulus p. It must be a prime number such that q divides (p-1). P is used in verification of shares.
     * @param q The modulus q. It must be a prime number fitting the requirement for p. This is the primary prime used in the sharing of the secret.
     * @param g G is a member of the subgroup generated by q in the field Z[p]. G is used in the verification of shares.
     * @param h H must also be in the same subgroup as g. It is used in verification of shares.
     * @param r The share number (as a <CODE>BigInteger</CODE>) of the share that needs to be recovered (during the sub-share operation). If null, then split will assume that no share needs to be recovered.
     * @throws ShareUtilException An indication of a problem with the underlying share functions.
     * @throws SplitException Indicates an error in the split process. Additional information is available from the exception message.
     */
    public static Share[] split(BigInteger secret, int n, int k, int numBits, BigInteger p, BigInteger q, BigInteger g, BigInteger h, BigInteger[] commitmentValues, BigInteger r)
    throws SplitException, ShareUtilException {
        //r is the recovery number - the polynomials will be modified such that f(r) = g(r) = 0;
        if(commitmentValues == null) {
            throw new SplitException("The commitmentValues are null");
        }
        if(secret == null) {
            throw new SplitException("The secret is null");
        }
        if(p == null) {
            throw new SplitException("p is null");
        }
        if(q == null) {
            throw new SplitException("q is null");
        }
        if(g == null) {
            throw new SplitException("g is null");
        }
        if(h == null) {
            throw new SplitException("h is null");
        }
        if(n < k) {
            throw new SplitException("n is less than k");
        }
        if(k != commitmentValues.length) {
            throw new SplitException("Incorrect length for commitment values array.");
        }
        if(!BigInteger.ONE.equals(p.mod(q))) {
            throw new SplitException("q must divide p-1");
        }
        if(!BigInteger.ONE.equals(g.modPow(q, p))) {
            throw new SplitException("g^q must be congruent to 1 mod p");
        }
        if(!BigInteger.ONE.equals(h.modPow(q, p))) {
            throw new SplitException("h^q must be congruent to 1 mod p");
        }
        
        SecureRandom sr = new SecureRandom();
        
        BigInteger[] polyF = new BigInteger[k];
        BigInteger[] polyG = new BigInteger[k];
        
        polyF[0] = secret;
        
        Share[] returnShareArray = new Share[n];
        
        if(!secret.equals(BigInteger.ZERO)) {
            //generate random commitment value for secret
            polyG[0] = new BigInteger(numBits, sr);
        } else {
            //set commitment value to zero
            polyG[0] = BigInteger.ZERO;
        }
        
        int i;
        
        //create the polynomials
        for(i = 1;i < k;i++) {
            polyF[i] = new BigInteger(numBits, sr);
            polyF[i] = polyF[i].mod(q);
            polyG[i] = new BigInteger(numBits, sr);
            polyG[i] = polyG[i].mod(q);
        }
        
        //Debugging
/*        System.out.print("F(x) = ");
        for(i = 0;i < k;i++) {
            System.out.print(polyF[i].toString());
            if(i != 0) {
                System.out.print("x^" + i);
            }
            if(i != k - 1) {
                System.out.print(" + ");
            }
        }
        System.out.println();
        System.out.print("G(x) = ");
        for(i = 0;i < k;i++) {
            System.out.print(polyG[i].toString());
            if(i != 0) {
                System.out.print("x^" + i);
            }
            if(i != k - 1) {
                System.out.print(" + ");
            }
        }
        System.out.println();
 */
        //end Debugging
        
        //correct the appropriate shares if r != null
        if(r != null) {
            BigInteger tempY = ShareUtil.evaluatePoly(polyF,q,r.intValue());
            tempY = tempY.negate();
            tempY = tempY.add(q);
            polyF[0] = tempY.mod(q);
            
            tempY = ShareUtil.evaluatePoly(polyG,q,r.intValue());
            tempY = tempY.negate();
            tempY = tempY.add(q);
            polyG[0] = tempY.mod(q);
        }
        
        //generate the commitment Values
        //System.out.println("generating the commitment values");
        for(i = 0;i < k;i++) {
            if(polyF[i] == null) {
                System.out.println("polyF[" + i + "] is null");
            }
            commitmentValues[i] = ShareUtil.calculatePublicVersion(polyF[i], polyG[i], p, g, h);
        }
        
        BigInteger[] secretShares = new BigInteger[n];
        BigInteger[] commitShares = new BigInteger[n];
        BigInteger[] publicShares = new BigInteger[n];
        
        //creates the values for the shares
        //System.out.println("creating the values for the shares");
        for(i = 1;i <= n;i++) {
            secretShares[i-1] = ShareUtil.evaluatePoly(polyF,q,i);
            commitShares[i-1] = ShareUtil.evaluatePoly(polyG,q,i);
            if(secretShares[i-1] == null) {
                System.out.println("secretShare " + (i-1) + " is null");
            }
            publicShares[i-1] = ShareUtil.calculatePublicVersion(secretShares[i-1], commitShares[i-1], p, g, h);
        }
        
        //creates the shares where i is the share number (array index is i-1)
        try {
            for(i = 1;i <= n;i++) {
                returnShareArray[i-1] = new Share(i, secretShares[i-1], commitShares[i-1], p, q, g, h, publicShares, k);
            }
        } catch (ShareException e) {
            throw new SplitException("Error creating share " + i + e.getMessage());
        }
        
        return returnShareArray;
    }
    
    /** <P>A recovery mode, it is used to indicate the recovery of the secret.</P>
     */
    public static final int RECOVER_SECRET = 0;
    /** <P>A recovery mode, it is used to indicate the recovery of the commitment value.</P>
     */
    public static final int RECOVER_COMMIT = 1;
    
    /** <P>This method recovers the secret from an array of shares. The array of shares does not need to be in any particular order and all of the shares will be verified before the recovery.</P>
     * <P><CODE>shares[0]</CODE> is assumed to be the local share and therefore the combiner knows that it is correct. In the event that the combiner knows that the local share is incorrect, then it should replace <CODE>share[0]</CODE> with a previously verified share. The first share will be used as a correct basis for the other shares.</P>
     * <P>The recovery method can recover either the secret or the commitment value for the secret. The mode indicates which of the two to recover.</P>
     * <P>If the goal is to recover a specific share, it needs to be done at that x value of the function rather than zero, the x-value of the secret/commitment value. The x-value is represented as a <CODE>BigInteger</CODE> and should be null to indicate zero.</P>
     * <P>Fix: verion .5 specified the necessity of having share 0 not null.</P>
     * @return The recovered secret or commitment value.
     * @param recoverAtX The <CODE>BigInteger</CODE> representation of the desired x value of the share to recover. Use null for zero.
     * @param mode Use one of the two recovery modes provided as static members of this class.
     * @param shares The array of shares for recovery. <CODE>Share[0]</CODE> is assumed to be correct.
     * @throws ShareException Indication of an error with the argument to {@link Share#verifyShare(Share) verifyShare}.
     * @throws RecoverException Indication of an error with the inputs.
     * @throws ShareUtilException Indication of an error in recovering the secret.
     */ 
    public static BigInteger recover(Share[] shares, BigInteger recoverAtX, int mode) 
    throws RecoverException, ShareUtilException, ShareException {
        int i;
        int j = 0;
        
        if(shares[0] == null) {
            throw new RecoverException("Share 0 is null");
        }
        
        BigInteger[] secretShares = new BigInteger[shares[0].getK()];
        BigInteger[] shareNum = new BigInteger[secretShares.length];
        int numBadShares = 0;
        
        //choses the appropriate secret shares to restore
        for(i = 0;i < shares.length && j < secretShares.length;i++) {
            if(i != 0) {
                if(shares[i] == null) {
                    numBadShares++;
                    continue;
                }
                if(!shares[0].verifyShare(shares[i])) {
                    numBadShares++;
                    continue;
                }
            }
            switch(mode) {
                case RECOVER_COMMIT:
                    secretShares[j] = shares[i].getSecretCommitment();
                    break;
                case RECOVER_SECRET:
                    secretShares[j] = shares[i].getSecretShare();
                    break;
                default:
                    throw new RecoverException("Invalid mode");
            }
            shareNum[j] = new BigInteger(Integer.toString(shares[i].getShareNumber()));
            j++;
        }
        
        if(shares.length - numBadShares < secretShares.length) {
            throw new RecoverException("Insufficient number of valid shares to reconstruct secret");
        }
        
        //Debugging
/*        System.out.println("Recover values");
        for(i = 0;i < secretShares.length;i++) {
            System.out.println("\tShare " + shareNum[i] + " - " + secretShares[i].toString());
        }
 */
        //end debugging
        
        BigInteger prodTermCum;
        BigInteger prodTerm;
        BigInteger prodXjSubXi;
        BigInteger prodXjSubXiInverse;
        BigInteger prodXjSubX;
        
        BigInteger sumTerm;
        BigInteger sumTermCum = BigInteger.ZERO;
        
        //computes the LaGrange Interpolation of the shares (only using the valid ones)
        for(i = 0;i < secretShares.length;i++) {
            prodTermCum = BigInteger.ONE;
            for(j = 0;j < secretShares.length;j++) {
                if(i == j) {
                    continue;
                }
                //(X[j] - X[i])^-1
                prodXjSubXi = shareNum[j].subtract(shareNum[i]);
                prodXjSubXi = prodXjSubXi.add(shares[0].getQ());
                prodXjSubXi = prodXjSubXi.mod(shares[0].getQ());
                prodXjSubXiInverse = prodXjSubXi.modInverse(shares[0].getQ());
                
                //X[j] - recoverAtX
                if(recoverAtX == null) {
                    prodXjSubX = shareNum[j].add(BigInteger.ZERO);
                } else {
                    prodXjSubX = shareNum[j].subtract(recoverAtX);
                    prodXjSubX = prodXjSubX.mod(shares[0].getQ());
                }
                
                //(X[j] - recoverAtX) * (X[j] - X[i])^-1
                prodTerm = prodXjSubX.multiply(prodXjSubXiInverse);
                prodTerm = prodTerm.mod(shares[0].getQ());
                
                //multiply all portions of the weight except for the share
                prodTermCum = prodTermCum.multiply(prodTerm);
                prodTermCum = prodTermCum.mod(shares[0].getQ());
            }
            sumTerm = secretShares[i].multiply(prodTermCum);
            sumTerm = sumTerm.mod(shares[0].getQ());
            
            sumTermCum = sumTermCum.add(sumTerm);
            sumTermCum = sumTermCum.add(shares[0].getQ());
            sumTermCum = sumTermCum.mod(shares[0].getQ());
        }
        
        return sumTermCum;
    }
    
    /** <p>Random generates the values: p, q, g and h.</P>
     * <p>The values are returned in an array:
     * <ul>
     *    <LI><CODE>return[0] = p;</CODE></LI>
     *    <LI><CODE>return[1] = q;</CODE></LI>
     *    <LI><CODE>return[2] = g;</CODE></LI>
     *    <LI><CODE>return[3] = h;</CODE></LI>
     * </UL>
     * <P>The method was updated in version .4 to deal with the fact that the <CODE>BigInteger</CODE> class doesn't create BigIntegers of a certain length, just a random number between 2<SUP>numBits</SUP> to 1.</P>
     * @return The values p, q, g and h in a <CODE>BigInteger</CODE> Array.
     * @param sr The source of randomness for the values.
     * @param numBits The number of bits for the values.
     */
    public static BigInteger[] generateValues(SecureRandom sr, int numBits) {
        BigInteger[] returnArray = new BigInteger[4];
        BigInteger qMult;
        BigInteger[] multArray = new BigInteger[3];
        int i;
        
        multArray[0] = BigInteger.ONE;
        multArray[1] = new BigInteger(Integer.toString(4));
        multArray[2] = new BigInteger(Integer.toString(8));
        
        boolean pFound = false;
        
        do {
            //first chose q
            returnArray[1] = BigInteger.probablePrime(numBits + 2, sr);
            
            if(returnArray[1].bitLength() < (numBits + 1)) {
                continue;
            }
        
            //find p such that p = 1 mod q
            //try 2q+1, 4q+1, and 8q+1 for p
            for(i = 0;i < 3;i++) {
                qMult = returnArray[1].multiply(multArray[i]);
                returnArray[0] = BigInteger.ONE.add(qMult);
                if(returnArray[0].isProbablePrime(10)) {
                    //probability that p is a prime: 0.9990234375
                    pFound = true;
                    break;
                }
            }
            
            //returnArray[0] = BigInteger.probablePrime(numBits + 4, sr);
        } while(!pFound);
        
        //find g such that g^q = 1 mod p
        do {
            returnArray[2] = new BigInteger(numBits, sr);
        } while(!BigInteger.ONE.equals(returnArray[2].modPow(returnArray[1], returnArray[0])) || (returnArray[2].bitLength() < numBits));
        
        // find h such that h^q = 1 mod p
        do {
            returnArray[3] = new BigInteger(numBits, sr);
        } while(!BigInteger.ONE.equals(returnArray[3].modPow(returnArray[1], returnArray[0])) || (returnArray[3].bitLength() < numBits));
        
        return returnArray;
    }
    
    /** <P>This function calculates the public version of a <CODE>Share</CODE>.</P>
     * @return A <CODE>BigInteger</CODE> corresponding to the public version of the share values.
     * @param secretShare A <CODE>BigInteger</CODE> with the value of the share's secret share.
     * @param secretCommitment A <CODE>BigInteger</CODE> with the value of the share's secret commitment share.
     * @param p A <CODE>BigInteger</CODE> with the same value as the prime modulus p used by the share.
     * @param g A <CODE>BigInteger</CODE> with the same value as the group member g used by the share.
     * @param h A <CODE>BigInteger</CODE> with the same value as the group member h use by the share.
     * @throws ShareUtilException Indicates error during the calculation of the public share; including parameter errors.
     */
    public static BigInteger calculatePublicVersion(BigInteger secretShare, BigInteger secretCommitment, BigInteger p, BigInteger g, BigInteger h) 
    throws ShareUtilException {
        if(secretShare == null) {
            throw new ShareUtilException("secretShare is null");
        }
        if(secretCommitment == null) {
            throw new ShareUtilException("secretCommitment is null");
        }
        if(p == null) {
            throw new ShareUtilException("p is null");
        }
        if(g == null) {
            throw new ShareUtilException("g is null");
        }
        if(h == null) {
            throw new ShareUtilException("h is null");
        }
        BigInteger gRaisedToS = g.modPow(secretShare, p);
        BigInteger hRaisedToT = h.modPow(secretCommitment, p);
        BigInteger calculatedPublicShare = gRaisedToS.multiply(hRaisedToT);
        return calculatedPublicShare.mod(p);
    }
    
    
    /** <P>This function is crucial to the splitting function. It provides a simple way to evaluate the polynomial at each separate <CODE>shareNumber</CODE>.</P>
     * @return A <CODE>BigInteger</CODE> equal to poly(x) mod q
     * @param poly An array of <CODE>BigIntegers</CODE> cooresponding to the coefficients of a <CODE>k-1</CODE> degree polynomial.
     * @param q The prime modulus q used in the splitting exercise.
     * @param x The evaluation parameter for the polynomial.
     * @throws ShareUtilException An error with the parameters has occurred.
     * @throws ArithmeticException There has been an arithmetic error in the evaluation of the polynomial.
     */    
    public static BigInteger evaluatePoly(BigInteger[] poly, BigInteger q, int x) throws ShareUtilException, ArithmeticException {
        if(poly == null) {
            throw new ShareUtilException("Polynomial not initialized");
        }
        
        BigInteger result = BigInteger.ZERO;
        BigInteger xToExponent;
        BigInteger term;
        
        for(int i = 0;i < poly.length;i++) {
            xToExponent = new BigInteger(Integer.toString((int)Math.pow((double)x,(double)i)));
            term = poly[i].multiply(xToExponent);
            term = term.mod(q);
            result = result.add(term);
            result = result.mod(q);
        }
        
        return result;
    }     
}
