/*
 * Client.java
 *
 * Created on December 8, 2002, 10:08 PM
 */

package demo;

import Secret_Share.*;

import java.io.*;
import java.util.*;
import java.math.BigInteger;
import java.security.*;

//graphics files so the client can make its window
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** <P>The client class is a thread that sports a <CODE>JFrame</CODE> window. The purpose of the thread is to act like an independent client, without actually being a truly independent client. The "dealer" remains in control of each client, but only communicates with each client by creating the same kinds of messages that would be used by networked clients.</P>
 * <P>The only <CODE>JFrame</CODE> component is the <CODE>ClientDisplay</CODE> class, an extension of <CODE>JPanel</CODE>. All GUI changes are made via the <CODE>ClientDisplay</CODE> class. There is also a hook for the one button, "corrupt", in the display to the <CODE>share.corrupt()</CODE> method.</P>
 * @author Brice Daniels
 * @version .4
 */
public class Client extends Thread {
    private int shareNumber;
    private int numBits = 0;
    private int round = 0;
    private BigInteger originalSecret = null;
    private BigInteger originalCommit = null;
    private Share local = null;
    private JFrame clientWindow = null;
    private RSACryptoManager cryptoManager = null;
    private ClientDisplay display = null;
    private BigInteger[] recoveryPublicVersion = null;
    
    /** Creates a new instance of Client
     * @param numBits The minimum number of bits in the <CODE>BigInteger</CODE>.
     * @param shareNumber The share number of the current client.
     * @param privateKey The RSA private key for the current client.
     * @param publicKeys The array of RSA public keys for all of the clients.
     */
    public Client(int numBits, int shareNumber, PrivateKey privateKey, PublicKey[] publicKeys) {
        this.numBits = numBits;
        this.shareNumber = shareNumber;
        this.cryptoManager = new RSACryptoManager(this, privateKey, publicKeys);
        clientWindow = new JFrame("Client Share " + this.shareNumber);
        this.display = new ClientDisplay(this);
    }
    
    /** <P>Returns a reference to the client's <CODE>RSACryptoManager</CODE>.</P>
     * @return The <CODE>RSACryptoManager</CODE> for the current client.
     */    
    protected RSACryptoManager getCryptoManager() {
        return this.cryptoManager;
    }
    
    /** <P>Returns the share number of the current client. As it is public information, this is available outside of the package.</P>
     * @return The share number as an <CODE>int</CODE>.
     */    
    public int getShareNumber() {
        return this.shareNumber;
    }
    
    /** <P>The current round number. This is needed by the dealer and a few other methods. The round number is maintained by the client and is not part of the secret share.</P>
     * @return The round number as an <CODE>int</CODE>.
     */    
    protected int getRound() {
        return this.round;
    }
    
    /** <P>Increments the current round number for this share.</P>
     */    
    protected void incrementRound() {
        this.round++;
    }
    
    /** <P>Overrides the run method of class Thread. It starts up the GUI component of the client.</P>
     */    
    public void run() {
        Container clientWindowContainer = clientWindow.getContentPane();
        clientWindowContainer.setLayout(new FlowLayout());
        clientWindowContainer.add(this.display);
        clientWindow.pack();
        clientWindow.show();
    }
    
    /** <P>This method is used by the dealer to initially give the client its share. The share is in a "message" for and includes a signature.</P>
     * <P>The first task of the client is to verify the signature and then decrypt the secret share and commitment values.</P>
     * <P>Provided all goes well with the verification and the decryption, the share is added to the client.</P>
     * @param input A <CODE>byte</CODE> encoded version of the message from the dealer.
     * @param senderShareNumber The share number of the dealer.
     * @throws ClientException Indication of an error processing the message.
     */    
    protected void receiveShare(byte[] input, int senderShareNumber) throws ClientException {
//        System.out.println("receiveShare - start");
        if(input == null) {
            throw new ClientException("Received Share input is null");
        }
        LinkedList dataStorage = this.processMessage(input, senderShareNumber);
        
        this.round = ((Integer) dataStorage.removeFirst()).intValue();
        byte[] secretData = (byte[]) dataStorage.removeFirst();
        BigInteger[] eArray = (BigInteger[]) dataStorage.removeFirst();
        BigInteger[] y = (BigInteger[]) dataStorage.removeFirst();
        BigInteger p = (BigInteger) dataStorage.removeFirst();
        BigInteger q = (BigInteger) dataStorage.removeFirst();
        BigInteger g = (BigInteger) dataStorage.removeFirst();
        BigInteger h = (BigInteger) dataStorage.removeFirst();
        int k = ((Integer) dataStorage.removeFirst()).intValue();
        
        BigInteger[] secrets = null;
        
        try {
            secrets = this.cryptoManager.decrypt(secretData);
        } catch (CryptoException e) {
            display.displayError("Receive Share","Error decrypting secret\n" + e.toString());
            throw new ClientException("Decryption Error\n" + e.toString());
        } catch (MalformedCryptoException e) {
            display.displayError("Receive Share","Error decrypting secret\n" + e.toString());
            throw new ClientException("Decryption Error\n" + e.toString());
        }
        
        try {
            this.local = new Share(this.shareNumber,secrets[0],secrets[1],p,q,g,h,y,k);
        } catch (SecretShareException e) {
            display.displayError("Receive Share","Error creating share\n" + e.toString());
            throw new ClientException("Error creating share\n" + e.toString());
        }
        try {
            if(!this.local.verify(eArray)) {
                display.displayError("Receive Share","The share is invalid");
                throw new ClientException("The share is invalid");
            }
        } catch (SecretShareException e) {
            display.displayError("Receive Share","Error validating share\n" + e.toString());
            throw new ClientException("Error validating share\n" + e.toString());
        }
        this.originalSecret = this.local.getSecretShare();
        this.originalCommit = this.local.getSecretCommitment();
        this.display.updateDisplay();
//        System.out.println("receive share - end");
    }
    
    /** <P>Instructs the client to corrupt its share.</P>
     * @throws ClientException Indication of an error corrupting the client.
     */    
    protected void corruptShare() throws ClientException {
        if(local == null) {
            throw new ClientException("The local share is null.");
        }
        this.local.corruptShare(numBits);
        this.display.updateDisplay();
    }
    
    /** <P>Gets the secret as a string for display on the client window.</P>
     * @return A string with the secret or zero if the secret is null.
     */    
    protected String getSecret() {
        if(this.local != null) {
            return this.local.getSecretShare().toString();
        } else {
            return BigInteger.ZERO.toString();
        }
    }
    
    /** <P>Gets the commitment value as a string for display on the client window.</P>
     * @return A string with the commitment value or zero if the commitment value is null.
     */    
    protected String getCommitment() {
        if(this.local != null) {
            return this.local.getSecretCommitment().toString();
        } else {
            return BigInteger.ZERO.toString();
        }
    }
    
    /** <P>Gets the original secret as a string for display on the client window.</P>
     * @return A string with original secret or zero if the original secret is null.
     */    
    protected String getOriginalSecret() {
        if(this.originalSecret != null) {
            return this.originalSecret.toString();
        } else {
            return BigInteger.ZERO.toString();
        }
    }
    
    /** <P>Gets the original commitment value as a string for display on the client window.</P>
     * @return A string with the original commitment value or zero if the original commitment value is null.
     */    
    protected String getOriginalCommitment() {
        if(this.originalCommit != null) {
            return this.originalCommit.toString();
        } else {
            return BigInteger.ZERO.toString();
        }
    }
    
    /** <P>Takes the data for the message as a <CODE>LinkedList</CODE> and turns it into a <CODE>byte</CODE> array along with a digital signature for the data bytes.</P>
     * @param data A <CODE>LinkedList</CODE> containing all of the data for the message.
     * @throws ClientException Indication of an error creating the message.
     * @return A <CODE>byte</CODE> array with the serialized data and a digital signature.
     */    
    protected byte[] genMessage(LinkedList data) throws ClientException {
//        System.out.println("Start genMessage");
        ByteArrayOutputStream bOut = new ByteArrayOutputStream();
        ObjectOutputStream out = null;
        try {
            out = new ObjectOutputStream(bOut);

            out.writeObject(data);
            out.flush();
        } catch (IOException e) {
            throw new ClientException("Error creating data portion of the message\n" + e.toString());
        }
        
        byte[] dataByte = bOut.toByteArray();
        
//        System.out.println("End genMesage - still signing");
        
        try {
            return cryptoManager.sign(dataByte);
        } catch (CryptoException e) {
            throw new ClientException("Error generating message: \n" + e.toString());
        }
    }
    
    /** <P>Performs the opposite operation as {@link Client#genMessage(LinkedList) genMessage()}. It takes the <CODE>byte</CODE> array and the share number of the sender and verifies the digital signature and then returns the data as a <CODE>LinkedList</CODE>.</P>
     * @param message The serialized message created by {@link Client#genMessage(LinkedList) genMessage()}
     * @param senderShareNumber The share number of the sender. Used to identify the proper public key to use.
     * @throws ClientException Indication of an error processing the message.
     * @return A <CODE>LinkedList</CODE> with the data from the message.
     */    
    protected LinkedList processMessage(byte[] message, int senderShareNumber) throws ClientException {
        if(message == null) {
            throw new ClientException("Message from client " + senderShareNumber + " is null");
        }
        byte[] data = null;
        try {
            data = cryptoManager.validate(message,senderShareNumber);
        } catch (CryptoException e) {
            throw new ClientException("Error validating input\n" + e.toString());
        } catch (MalformedCryptoException e) {
            display.displayError("Receive Share","Error receiving share\n" + e.toString());
            throw new ClientException("Corrupt Data Received\n" + e.toString());
        }
        
        ByteArrayInputStream bIn = new ByteArrayInputStream(data);
        ObjectInputStream in = null;
        LinkedList dataStorage = null;
        
        try {
            in = new ObjectInputStream(bIn);
            dataStorage = (LinkedList) in.readObject();
            in.close();
        } catch (Exception e) {
            display.displayError("Receive Share","Error recovering share from message\n" + e.toString());
            throw new ClientException("IO Errors\n" + e.toString());
        }
        
        return dataStorage;
    }
    
    /** <P>Validate the client's share.</P>
     * @return True if the client is valid.
     */    
    protected boolean validate() {
        boolean validShare = false;
        try {
            validShare = this.local.selfVerify();
        } catch (SecretShareException e) {
            display.displayError("Validate Share Error","Error validating share\n" + e.toString());
        }
        return validShare;
    }
    
    /** <P>Performs the first of two steps for share recovery for valid shares.</P>
     * <P>This method creates a message that contains the subshares such that f(r) = g(r) = 0.</P>
     * @param recoverShareNumber The share number of client that you are trying to recover.
     * @throws ClientException Indication of an error creating the message.
     * @return An array of digitally signed messages, one for each client.
     */    
    protected byte[][] recoverShareInit(int recoverShareNumber) throws ClientException {
        BigInteger[] recoveryEArray = new BigInteger[this.local.getK()];
        Share[] recoverySubShares = null;
        try {
            recoverySubShares = ShareUtil.split(BigInteger.ZERO,this.local.getN(),this.local.getK(),this.numBits,this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),recoveryEArray,new BigInteger(Integer.toString(recoverShareNumber)));
        } catch (SecretShareException e) {
            this.display.displayError("Recover Share Init","Error initiating recovery for share " + recoverShareNumber + "\n" + e.toString());
            throw new ClientException("Error initiating recovery for share " + recoverShareNumber + "\n" + e.toString());
        }
        
        LinkedList data = new LinkedList();
        byte[][] messages = new byte[this.local.getN()][];
        
        for(int i = 0;i < this.local.getN();i++) {
            data.clear();
            data.addLast(new Integer(recoverShareNumber));
            try {
                data.addLast(this.cryptoManager.encrypt(recoverySubShares[i],i+1));
            } catch (CryptoException e) {
                this.display.displayError("Recover Share Init","Error initiating recovery for share " + recoverShareNumber + "\n" + e.toString());
                throw new ClientException("Error initiating recovery for share " + recoverShareNumber + "\n" + e.toString());
            }
            data.addLast(recoveryEArray);
            data.addLast(recoverySubShares[0].getAllPublicVersion());
            
            messages[i] = this.genMessage(data);
        }
        
        return messages;
    }
    
    /** <P>The second of two methods used by valid shares in share recovery. It creates shares of the bad share.</P>
     * <P>The function creates a message for the invalid share that contains the share of the invalid share.</P>
     * @param messages The messages from the other valid servers.
     * @throws ClientException An indication of an error creating the message for the invalid client.
     * @return A digitally signed message for the invalid client.
     */    
    protected byte[] recoverShareFinal(byte[][] messages) throws ClientException {
        int i;
        Share recoverTest = null;
        Share recoveryShare = null;
        
        try {
            recoveryShare = this.local.copy();
        } catch (SecretShareException e) {
            this.display.displayError("Recover Share Processing","Error creating share copy\n" + e.toString());
            throw new ClientException("Error creating share copy\n" + e.toString());
        }
        
        LinkedList data = null;
        BigInteger[] recoveryEArray = null;
        Share recoverySubShare = null;
        BigInteger[] subSecret = null;
        BigInteger[] subY = null;
        Integer badShare = null;
        
        for(i = 0;i < messages.length;i++) {
            if(messages[i] == null) {
                continue;
            }
            
            data = this.processMessage(messages[i], i+1);
            badShare = (Integer) data.removeFirst();
            try {
                subSecret = this.cryptoManager.decrypt((byte[]) data.removeFirst());
            } catch (MalformedCryptoException e) {
                this.display.displayError("Recover Share Processing","Invalid encryption of sub share " + (i+1) + "\n" + e.toString());
                throw new ClientException("Invalid encryption of sub share " + (i+1) + "\n" + e.toString());
            } catch (CryptoException e) {
                this.display.displayError("Recover Share Processing","Error decryption of sub share " + (i+1) + "\n" + e.toString());
                throw new ClientException("Error decryption of sub share " + (i+1) + "\n" + e.toString());
            }
            recoveryEArray = (BigInteger[]) data.removeFirst();
            subY = (BigInteger[]) data.removeFirst();
            
            
            try {
                recoverTest = new Share(badShare.intValue(),BigInteger.ZERO,BigInteger.ZERO,this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),subY,this.local.getK());
            } catch (ShareException e) {
                this.display.displayError("Recover Share Processing","Error creating recover test share\n" + e.toString());
                throw new ClientException("Error creating recover test share\n" + e.toString());
            }
            
            try {
                recoverySubShare = new Share(this.shareNumber,subSecret[0],subSecret[1],this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),subY,this.local.getK());
            } catch (SecretShareException e) {
                this.display.displayError("Recover Share Processing","Error creating sub share\n" + e.toString());
                throw new ClientException("Error creating sub share\n" + e.toString());
            }
            try {
                if(!recoverTest.verify(recoveryEArray)) {
                    this.display.displayError("Recover Share Processing","Recover shares generated by " + (i+1) + " failed recover test");
                    continue;
                }
            } catch (SecretShareException e) {
                this.display.displayError("Recover Share Processing","Error verifying recoveryEArray " + (i+1) + "\n" + e.toString());
                throw new ClientException("Error verifying recoveryEArray " + (i+1) + "\n" + e.toString());
            }
            
            try {
                if(!recoverySubShare.verify(recoveryEArray)) {
                    this.display.displayError("Recover Share Processing","Client " + this.shareNumber + ":\nRecover subshare " + (i+1) + " failed verify test");
                    continue;
                }
            } catch (SecretShareException e) {
                this.display.displayError("Recover Share Processing","Error verifying recoverySubShare " + (i+1) + "\n" + e.toString());
                throw new ClientException("Error verifying recoverySubShare " + (i+1) + "\n" + e.toString());
            }
            try {
                recoveryShare.add(recoverySubShare,recoveryEArray);
            } catch (SecretShareException e) {
                this.display.displayError("Recover Share Processing","Error adding subshare " + (i+1) + "\n" + e.toString());
                throw new ClientException("Error adding subshare " + (i+1) + "\n" + e.toString());
            }
        }
        
        data.clear();
        data.addLast(new Integer(this.round));
        data.addLast(new Integer(recoveryShare.getShareNumber()));
        try {
            data.addLast(this.cryptoManager.encrypt(recoveryShare,badShare.intValue()));
        } catch (CryptoException e) {
            this.display.displayError("Recover Share Processing","Error encrypting recovery share\n" + e.toString());
            throw new ClientException("Error encrypting recovery share\n" + e.toString());
        }
        
        return this.genMessage(data);
    }
    
    /** <P>The first of two methods used by invalid clients in share recovery.</P>
     * <P>This method looks at all of the messages from the valid clients and uses the commitment arrays to create public versions for the shares that it will receive later.</P>
     * @param messages The messages from the valid clients.
     * @throws ClientException An indication of an error gathering the commitment array values.
     */    
    protected void recoverShareListen(byte[][] messages) throws ClientException {
        if(messages == null) {
            throw new ClientException("Messages for client " + this.shareNumber + " are null");
        }
        LinkedList data = null;
        BigInteger[] subEArray = null;
        BigInteger exponent = null;
        BigInteger term = null;
        this.recoveryPublicVersion = this.local.getAllPublicVersion();
        for(int j = 0;j < messages.length;j++) {
            if(messages[j] == null) {
                continue; // skips over the bad clients, since they don't send messages
                          // this will not work in a live environment, as it opens holes ...
            }
            data = this.processMessage(messages[j], j+1);
            data.removeLast();
            subEArray = (BigInteger[]) data.removeLast();
            if(subEArray.length != this.local.getK()) {
                this.display.displayError("Recover Share Processing","Message from client " + (j+1) + " has the wrong EArray length.");
                throw new ClientException("Message from client " + (j+1) + " has the wrong EArray length.");
            }
            for(int i = 0;i < this.recoveryPublicVersion.length;i++) {
                for(int m = 0;m < subEArray.length;m++) {
                    exponent = new BigInteger(Integer.toString((int) Math.pow((double) (i+1), (double) m)));
                    term = subEArray[m].modPow(exponent,this.local.getP());
                    this.recoveryPublicVersion[i] = this.recoveryPublicVersion[i].multiply(term);
                    this.recoveryPublicVersion[i] = this.recoveryPublicVersion[i].mod(this.local.getP());
                }
            }
        }
        return;
    }
    
    /** <P>The second of two methods used by invalid clients in share recovery.</P>
     * <P>This method is used by the particular invalid client to recover its share from the shares given to it by the valid clients. It verifies those shares using the information that it collected in {@link Client#recoverShareListen(byte[][]) recoverShareListen()}.</P>
     * @param messages The messages from the valid clients containing the shares of this client's share.
     * @throws ClientException An indication of an error recovering the share.
     */    
    protected void recoverShare(byte[][] messages) throws ClientException {
        LinkedList data = null;
        int shareNumRemote = 0;
        int roundRemote = -1;
        BigInteger[] secrets = null;
        Share[] recoveryShares = new Share[this.local.getN()];
        for(int j = 0;j < messages.length;j++) {
            if(messages[j] == null) {
                continue;
            }
            data = this.processMessage(messages[j],j+1);
            
            roundRemote = ((Integer) data.removeFirst()).intValue();
            if(roundRemote != this.round) {
                this.round = roundRemote;
            }
            shareNumRemote = ((Integer) data.removeFirst()).intValue();
            try {
                secrets = this.cryptoManager.decrypt((byte[]) data.removeFirst());
                if(secrets.length != 2) {
                    continue;
                }
                recoveryShares[j] = new Share(shareNumRemote,secrets[0],secrets[1],this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),this.recoveryPublicVersion,this.local.getK());
            } catch (MalformedCryptoException e) {
                this.display.displayError("Recovering Share","Invalid encryption in share " + (j+1) + "\n" + e.toString());
                throw new ClientException("Invalid encryption in share " + (j+1) + "\n" + e.toString());
            } catch (CryptoException e) {
                this.display.displayError("Recovering Share","Error decryption in share " + (j+1) + "\n" + e.toString());
                throw new ClientException("Error decryption in share " + (j+1) + "\n" + e.toString());
            } catch (SecretShareException e) {
                this.display.displayError("Recovering Share","Error creating share " + (j+1) + "\n" + e.toString());
                throw new ClientException("Error creating share " + (j+1) + "\n" + e.toString());
            }
        }
        if(recoveryShares[0] == null) {
            // recoveryShares[0] is used as the basis for all recovery and therefore must not be null!
            // simply find another share that isn't null, that is correct and switch them
            for(int i = 1;i < recoveryShares.length;i++) {
                if(recoveryShares[i] != null) {
                    try {
                        if(recoveryShares[i].selfVerify()) {
                            recoveryShares[0] = recoveryShares[i];
                            recoveryShares[i] = null;
                            break;
                        }
                    } catch (SecretShareException e) {
                        continue;
                    }
                }
            }
            if(recoveryShares[0] == null) {
                this.display.displayError("Recovering Share","unable to replace share[0]");
                throw new ClientException("Unable to replace share[0]");
            }
        }
        BigInteger recoveredSecret = null;
        BigInteger recoveredCommit = null;
        BigInteger shareNum = new BigInteger(Integer.toString(this.shareNumber));
        try {
            recoveredSecret = ShareUtil.recover(recoveryShares,shareNum,ShareUtil.RECOVER_SECRET);
            recoveredCommit = ShareUtil.recover(recoveryShares,shareNum,ShareUtil.RECOVER_COMMIT);
            this.local = new Share(this.shareNumber,recoveredSecret,recoveredCommit,this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),this.local.getAllPublicVersion(),this.local.getK());
        } catch (SecretShareException e) {
            this.display.displayError("Recovering Share","Error recovering secrets\n" + e.toString());
            throw new ClientException("Error recovering secrets\n" + e.toString());
        }
        this.recoveryPublicVersion = null;
        try {
            if(!this.local.selfVerify()) {
                this.display.displayError("Recovering Share","Share " + this.shareNumber + " was not properly recovered.");
                throw new ClientException("Share " + this.shareNumber + " was not properly recovered.");
            }
        } catch (SecretShareException e) {
            this.display.displayError("Recovering Share","Share " + this.shareNumber + " encountered an error in verification.");
            throw new ClientException("Share " + this.shareNumber + " encountered an error in verification.");
        }
        this.display.updateDisplay();
    }
    
    /** <P>This method is used by clients to generate a message for the recoverer. The message contains the secret share and commitment value held by this client.</P>
     * @param recovererShareNumber The share number of the recoverer.
     * @throws ClientException An indication of an error creating the message.
     * @return A digitally signed message for the recoverer.
     */    
    protected byte[] genRecoverSecretMessage(int recovererShareNumber) throws ClientException {
        LinkedList data = new LinkedList();
        
        data.addLast(new Integer(this.shareNumber));
        data.addLast(new Integer(this.round));
        try {
            data.addLast(this.cryptoManager.encrypt(this.local,recovererShareNumber));
        } catch (CryptoException e) {
            this.display.displayError("Generating Recover Secret Message","Error encrypting secrets\n" + e.toString());
            throw new ClientException("Error encrypting secrets\n" + e.toString());
        }
        
        return this.genMessage(data);
    }
    
    /** <P>The method called by the recoverer to recover the secret from the data contained in the array of messages.</P>
     * @param messages An array of messages from all clients containing their secret share and commitment value.
     * @throws ClientException An indication of an error recovering the secret.
     * @return The secret as a <CODE>BigInteger</CODE>.
     */    
    protected BigInteger recoverSecret(byte[][] messages) throws ClientException {
        LinkedList data = null;
        BigInteger[] secrets = null;
        int shareNum;
        int roundNum;
        Share[] shares = new Share[this.local.getN()];
        int i;
        for(i = 0;i < messages.length;i++) {
            if(i == (this.shareNumber - 1)) {
                try {
                    shares[i] = this.local.copy();
                } catch (SecretShareException e) {
                    this.display.displayError("Recovering Secret","Error copying local share from client " + this.shareNumber);
                    throw new ClientException("Error copying local share from client " + this.shareNumber);
                }
                continue;
            }
            if(messages[i] == null) {
                continue;
            }
            data = this.processMessage(messages[i], i+1);
            
            shareNum = ((Integer) data.removeFirst()).intValue();
            roundNum = ((Integer) data.removeFirst()).intValue();
            try {
                secrets = this.cryptoManager.decrypt((byte[]) data.removeFirst());
            } catch (MalformedCryptoException e) {
                this.display.displayError("Recovering Secret","Message from client " + (i+1) + " is invalid\n" + e.toString());
                throw new ClientException("Message from client " + (i+1) + " is invalid\n" + e.toString());
            } catch (CryptoException e) {
                this.display.displayError("Recovering Secret","Error decrypting message from client " + (i+1) + "\n" + e.toString());
                throw new ClientException("Error decrypting message from client " + (i+1) + "\n" + e.toString());
            }
            
            if(this.round != roundNum) {
                continue;
            }
            
            if(secrets.length != 2) {
                continue;
            }
            
            try {
                shares[i] = new Share(shareNum,secrets[0],secrets[1],this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),this.local.getAllPublicVersion(),this.local.getK());
            } catch (SecretShareException e) {
                this.display.displayError("Recovering Secret","Error creating share for client " + (i+1) + "\n" + e.toString());
                throw new ClientException("Error creating share for client " + (i+1) + "\n" + e.toString());
            }
        }
        try {
            if(shares[0] == null || !shares[0].selfVerify()) {
                for(i = 1;i < shares.length;i++) {
                    if(shares[i] != null) {
                        try {
                            if(shares[i].selfVerify()) {
                                shares[0] = shares[i];
                                shares[i] = null;
                                break;
                            }
                        } catch (SecretShareException e) {
                            continue;
                        }
                    }
                }
                if(shares[0] == null) {
                    this.display.displayError("Recovering Secret","Unable to replace shares[0]");
                    throw new ClientException("Unable to replace shares[0]");
                }
            }
        } catch (SecretShareException e) {
            this.display.displayError("Recovering Secret","Error validating shares[0]");
            throw new ClientException("Error validating shares[0]");
        }
        
        try {
            return ShareUtil.recover(shares,null,ShareUtil.RECOVER_SECRET);
        } catch (SecretShareException e) {
            this.display.displayError("Recovering Secret","Error recovering secret\n" + e.toString());
            throw new ClientException("Error recovering secret\n" + e.toString());
        }
    }
    
    /** <P>The first of two functions involved in share renewal. It creates a share of zero and then a message to each client with that client's share of zero.</P>
     * @throws ClientException An indication of an error creating the array of messages.
     * @return An array of digitally signed messages containing the renewal information.
     */    
    protected byte[][] renewalInit() throws ClientException {
        try {
            if(!this.local.selfVerify()) {
                return null;
            }
        } catch (SecretShareException e) {
            this.display.displayError("Renewal Initialization","Error attempting to verify local share\n" + e.toString());
            throw new ClientException("Error attempting to verify local share\n" + e.toString());
        }
        BigInteger[] renewalEArray = new BigInteger[this.local.getK()];
        Share[] renewalSubShares = null;
        
        try {
            renewalSubShares = ShareUtil.split(BigInteger.ZERO,this.local.getN(), this.local.getK(), this.numBits, this.local.getP(), this.local.getQ(), this.local.getG(), this.local.getH(), renewalEArray);
        } catch (SecretShareException e) {
            this.display.displayError("Renewal Initialization","Error creating renewal subshares for client " + this.shareNumber + "\n" + e.toString());
            throw new ClientException("Error creating renewal subshares for client " + this.shareNumber + "\n" + e.toString());
        }
        
        byte[][] messages = new byte[this.local.getN()][];
        LinkedList data = new LinkedList();
        Integer roundObj = new Integer(this.round);
        BigInteger[] renewalPublicVersion = renewalSubShares[0].getAllPublicVersion();
        
        for(int i = 0;i < messages.length;i++) {
            data.clear();
            
            data.addLast(roundObj);
            try {
                data.addLast(this.cryptoManager.encrypt(renewalSubShares[i],i+1));
            } catch (CryptoException e) {
                this.display.displayError("Renewal Initialization","Error encrypting renewal subshares for client " + this.shareNumber + "\n" + e.toString());
                throw new ClientException("Error encrypting renewal subshares for client " + this.shareNumber + "\n" + e.toString());
            }
            data.addLast(renewalEArray);
            data.addLast(renewalPublicVersion);
            
            messages[i] = this.genMessage(data);
        }
        
        return messages;
    }
    
    /** <P>The second of two methods used for share renewal. It takes all messages from the other clients and adds each share of zero to its own share.</P>
     * @param messages An array of digitally signed messages from the other shares for this client.
     * @throws ClientException An indication of an error renewing the local share.
     */    
    protected void renewalFinal(byte[][] messages) throws ClientException {
        if(messages == null) {
            throw new ClientException("Messages is null");
        }
        
        boolean goodShare = true;
        
        try {
            if(!this.local.selfVerify()) {
                this.display.displayError("Share Renewal","local share " + this.shareNumber + " invalid, unable to renew.");
                goodShare = false;
            }
        } catch (SecretShareException e) {
            this.display.displayError("Share Renewal","Error verifying local share " + this.shareNumber);
            goodShare = false;
        }
        
        LinkedList data = null;
        int roundRemote = -1;
        BigInteger[] secrets = null;
        BigInteger[] renewalEArray = null;
        BigInteger[] renewalPublicVersion = null;
        Share renewalShare = null;
        
        for(int i = 0;i < messages.length;i++) {
            if(messages[i] == null) {
                continue;
            }
            
            data = this.processMessage(messages[i],i+1);
            
            roundRemote = ((Integer) data.removeFirst()).intValue();
            
            if(roundRemote != this.round) {
                continue;
            }
            
            try {
                secrets = this.cryptoManager.decrypt((byte[]) data.removeFirst());
            } catch (MalformedCryptoException e) {
                this.display.displayError("Renewal Processing","Invalid decryption\n" + e.toString());
                throw new ClientException("Invalid decryption\n" + e.toString());
            } catch (CryptoException e) {
                this.display.displayError("Renewal Processing","Error during decryption\n" + e.toString());
                throw new ClientException("Error during decryption\n" + e.toString());
            }
            if(secrets.length != 2) {
                continue;
            }
            renewalEArray = (BigInteger[]) data.removeFirst();
            if(renewalEArray.length != this.local.getK()) {
                continue;
            }
            renewalPublicVersion = (BigInteger[]) data.removeFirst();
            if(renewalPublicVersion.length != this.local.getN()) {
                continue;
            }
            
            try {
                renewalShare = new Share(this.shareNumber,secrets[0],secrets[1],this.local.getP(),this.local.getQ(),this.local.getG(),this.local.getH(),renewalPublicVersion,this.local.getK());
                if(!renewalShare.verify(renewalEArray)) {
                    continue;
                }
                this.local.add(renewalShare,renewalEArray);
            } catch (SecretShareException e) {
                this.display.displayError("Renewing Share","Share " + this.shareNumber + " encountered errors attempting to add subshare from " + (i+1) + "\n" + e.toString());
                throw new ClientException("Share " + this.shareNumber + " encountered errors attempting to add subshare from " + (i+1) + "\n" + e.toString());
            }
        }
        
        try {
            if(!this.local.selfVerify() && goodShare) {
                this.display.displayError("Renewing Share","Share " + this.shareNumber + " was not properly renewed.");
                throw new ClientException("Share " + this.shareNumber + " was not properly renewed.");
            }
        } catch (SecretShareException e) {
            this.display.displayError("Renewing Share","Share " + this.shareNumber + " encountered an error in verification.\n" + e.toString());
            throw new ClientException("Share " + this.shareNumber + " encountered an error in verification.\n" + e.toString());
        }
        this.round++;
        this.display.updateDisplay();
    }
}
