/*
 * Dealer.java
 *
 * Created on December 8, 2002, 9:36 PM
 */

package demo;

import Secret_Share.*;

import java.security.*;
import java.math.BigInteger;
import java.util.*;
import cryptix.provider.Cryptix;

/** <P>The dealer class runs the sharing operations loosely as the first client.</P>
 * @author Brice Daniels
 * @version .4
 */
public class Dealer {
    private Client[] clients = null;
    private SecureRandom sr = null;
    private BigInteger secret = null;
    private PublicKey[] publicKeys = null;
    private PrivateKey privateKey = null;
    private boolean[] badShares = null;
    
    /** Creates a new instance of Dealer */
    public Dealer() {
        //make sure that the Cryptix provider is registered
        Security.addProvider(new Cryptix());
//        Provider cryptix = Security.getProvider("Cryptix");
//        cryptix.list(System.out);
        //System.out.println(cryptix.getInfo());
        
        sr = new SecureRandom();
    }
    
    /** <P>Initializes the dealer and starts the secret sharing.</P>
     * @param n The number of clients.
     * @param k The threshold number.
     * @param secret The secret to share.
     * @param p The prime modulus p.
     * @param q The prime modulus q.
     * @param g The prime modulus g.
     * @param h The prime modulus h.
     * @param numBits The number of bits for all randomly generated values.
     * @throws DealerException An indication of error during initialization.
     */    
    protected void init(int n, int k, BigInteger secret, BigInteger p, BigInteger q, BigInteger g, BigInteger h, int numBits) throws DealerException{
        if(secret == null) {
//            do {
                secret = new BigInteger(numBits,sr);
//            } while (secret.bitLength() > numBits);
        }
        if(secret.bitLength() > numBits) {
            throw new DealerException("Secret is too long (" + secret.bitLength() +")");
        }
        if(p.bitLength() < (numBits + 3)) {
            throw new DealerException("P is too short (" + p.bitLength() +")");
        }
        if(q.bitLength() < (numBits + 1)) {
            throw new DealerException("Q is too short (" + q.bitLength() +")");
        }
        if(g.bitLength() < numBits) {
            throw new DealerException("g is too short (" + g.bitLength() +")");
        }
        if(h.bitLength() < numBits) {
            throw new DealerException("h is too short (" + h.bitLength() +")");
        }
        if(n < k) {
            throw new DealerException("n is less than k");
        }
        if(n <=0) {
            throw new DealerException("n must be positive and nonzero");
        }
        if(k <=0) {
            throw new DealerException("k must be positive and nonzero");
        }
        
        this.secret = secret;
        
        BigInteger[] eArray = new BigInteger[k];
        Share[] shares = null;
        try {
            shares = ShareUtil.split(secret,n,k,numBits,p,q,g,h,eArray);
        } catch (Exception e) {
            throw new DealerException(e.toString());
        }
        
        int i;
        
        KeyPair[] keys = new KeyPair[n];
        this.publicKeys = new PublicKey[n];
        for(i = 0;i < n;i++) {
            keys[i] = RSACryptoManager.generateKeyPair();
            this.publicKeys[i] = keys[i].getPublic();
        }
        this.privateKey = keys[0].getPrivate(); //the dealer is going to be clients[0]
        
        this.clients = new Client[n];
        LinkedList shareMessageData = new LinkedList();
        
        for(i = 0;i < n;i++) {
//            System.out.println("creating client " + (i+1));
            clients[i] = new Client(numBits,shares[i].getShareNumber(),keys[i].getPrivate(),this.publicKeys);
            clients[i].start();
            
            shareMessageData.clear();
            shareMessageData.addLast(new Integer(0)); //the current round (zero)
            try {
                shareMessageData.addLast(clients[0].getCryptoManager().encrypt(shares[i],i+1)); //the 
            } catch (CryptoException e) {
                throw new DealerException("Error encrypting shares\n" + e.toString());
            }
            shareMessageData.addLast(eArray);
            shareMessageData.addLast(shares[0].getAllPublicVersion());
            shareMessageData.addLast(p);
            shareMessageData.addLast(q);
            shareMessageData.addLast(g);
            shareMessageData.addLast(h);
            shareMessageData.addLast(new Integer(k));
            
//            System.out.println("Client " + (i+1) + " created, sending share");
            
            try {
                clients[i].receiveShare(clients[0].genMessage(shareMessageData),1);
            } catch (ClientException e) {
                throw new DealerException("Error sending share to client " + i + "\n" + e.toString());
            }
//            System.out.println("Client " + (i+1) + " received share");
        }
        
    }
    
    /** <P>Get the secret as a string.</P>
     * @return The secret as a string.
     */    
    protected String getSecret() {
        if(this.secret != null) {
            return this.secret.toString();
        } else {
            return "0";
        }
    }
    
    /** <P>The current round number.</P>
     * @return The current round number as a string.
     */    
    protected String getRound() {
        return Integer.toString(this.clients[0].getRound());
    }
    
    /** <P>Calls {@link Secret_Share.ShareUtil#generateValues(SecureRandom, int) ShareUtil.generateValues()}.</P>
     * @param numBits The number of bits to use in the random values.
     * @return The four public values as a <CODE>BigInteger</CODE> array.
     */    
    protected BigInteger[] generateValues(int numBits) {
        return ShareUtil.generateValues(sr,numBits);
    }
    
    /** <P>Attempts to recover shares found to be bad by {@link Dealer#validateShares() validateShares()}.</P>
     * @throws DealerException An indication of an error during recovery.
     * @return A string stating which clients were recovered. It is null if no clients were recovered.
     */    
    protected String recoverShares() throws DealerException {
        if(this.badShares == null) {
            throw new DealerException("The shares must be validated before recovery is attempted");
        }
        
        int i;
        int j;
        int m;
        byte[][][] messagesInit = new byte[this.clients.length][][];
        byte[][][] messagesOrdered = new byte[this.clients.length][][];
        byte[][] messagesFinal = new byte[this.clients.length][];
        String recoverMessage = null;
        
        for(i = 0;i < this.clients.length;i++) {
            if(!this.badShares[i]) {
                continue;
            }
            if(recoverMessage == null) {
                recoverMessage = Integer.toString(i+1);
            } else {
                recoverMessage += ", " + (i+1);
            }
            for(j = 0;j < this.clients.length;j++) {
                if(this.badShares[j]) {
                    continue;
                }
                try {
                    messagesInit[j] = this.clients[j].recoverShareInit(this.clients[i].getShareNumber());
                } catch (ClientException e) {
                    throw new DealerException("Error attempting to initialize recovery for share " + (j+1) + "\n" + e.toString());
                }
            }
            for(m = 0;m < this.clients.length;m++) {
                messagesOrdered[m] = new byte[this.clients.length][];
                for(j = 0;j < this.clients.length;j++) {
                    if(messagesInit[j] != null) {
                        messagesOrdered[m][j] = messagesInit[j][m];
                    }
                }
            }
            for(j = 0;j < this.clients.length;j++) {
                if(this.badShares[j]) {
                    try {
                        this.clients[j].recoverShareListen(messagesOrdered[j]);
                    } catch (ClientException e) {
                        throw new DealerException("Error listening by share " + (j+1) + "\n" + e.toString());
                    }
                    continue;
                }
                try {
                    messagesFinal[j] = this.clients[j].recoverShareFinal(messagesOrdered[j]);
                } catch (ClientException e) {
                    throw new DealerException("Error creating final message from share " + (j+1) + "\n" + e.toString());
                }
//                if(messagesFinal[j] == null) {
//                    System.out.println("Final message from client "+(j+1)+ " is null.");
//                }
            }
            try {
                this.clients[i].recoverShare(messagesFinal);
            } catch (ClientException e) {
                throw new DealerException("Error recovering share " + (i+1) + "\n" + e.toString());
            }
            this.badShares[i] = false;
        }
        this.badShares = null;
        return recoverMessage;
    }
    
    /** <P>Used to validate the clients. It creates a list of bad shares for {@link Dealer#recoverShares() recoverShares()}.</P>
     * @return The number of invalid clients.
     */    
    protected int validateShares() {
        this.badShares = new boolean[this.clients.length];
        int numBad = 0;
        for(int i = 0;i < clients.length;i++) {
            if(!clients[i].validate()) {
                this.badShares[i] = true;
                numBad++;
            } else {
                this.badShares[i] = false;
            }
        }
        return numBad;
    }
    
    /** <P>Takes the appropriate steps to recover the secret.</P>
     * @throws DealerException An indication of error during the attempt to recover the secret.
     * @return The secret as a <CODE>BigInteger</CODE>.
     */    
    protected BigInteger recoverSecret() throws DealerException {
        byte[][] messages = new byte[this.clients.length][];
        for(int i = 1;i < this.clients.length;i++) {
            try {
                messages[i] = this.clients[i].genRecoverSecretMessage(1);
            } catch (ClientException e) {
                throw new DealerException("Error creating recover secret message for client " + (i+1) + "\n" + e.toString());
            }
            if(messages[i] == null) {
                throw new DealerException("Recover secret message from client " + (i+1) + " is null");
            }
        }
        try {
            return this.clients[0].recoverSecret(messages);
        } catch (ClientException e) {
            throw new DealerException("Error recovering secret\n" + e.toString());
        }
    }
    
    /** <P>Attempts to renew the shares of all of the clients.</P>
     * @throws DealerException An indication of an error while renewing shares.
     */    
    protected void renewShares() throws DealerException {
        byte[][][] messages = new byte[this.clients.length][][];
        byte[][][] messagesOrdered = new byte[this.clients.length][][];
        int i;
        int j;
        int m;
        for(i = 0;i < messages.length;i++) {
            try {
                messages[i] = this.clients[i].renewalInit();
            } catch (ClientException e) {
                throw new DealerException("Error initializing renewal with client " + (i+1) + "\n" + e.toString());
            }
        }
        for(i = 0;i < this.clients.length;i++) {
            messagesOrdered[i] = new byte[this.clients.length][];
            for(j = 0;j < this.clients.length;j++) {
                if(messages[j] != null) {
                    messagesOrdered[i][j] = messages[j][i];
                }
            }
        }
        for(i = 0;i < this.clients.length;i++) {
            try {
                this.clients[i].renewalFinal(messagesOrdered[i]);
            } catch (ClientException e) {
                throw new DealerException("Error processing renewal with client " + (i+1) + "\n" + e.toString());
            }
        }
    }
}
