/*
 * RSACryptoManager.java
 *
 * Created on December 10, 2002, 2:29 PM
 */

package demo;

import Secret_Share.Share;

import java.io.*;
import java.util.*;
import java.math.BigInteger;

import java.security.SecureRandom;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import xjava.security.interfaces.*;
import xjava.security.*;
import java.security.*;
import cryptix.provider.rsa.*;

/** <P>A collection of purpose built methods for encrypting secret shares and commitment values and signing messages.</P>
 * <P>The RSA CryptoManager requires the cryptix provider, version 3.2 or later. In order to run the {@link ShareDemo ShareDemo}, you must have cryptix installed properly in your classpath or specified on the command line.</P>
 * @author Brice Daniels
 * @version .2
 * @see <A href="http://www.cryptix.org">Cryptix</A>
 */
public class RSACryptoManager {
    PrivateKey priv = null;
    PublicKey[] pubs = null;
    Client client = null;
    
    /** Creates a new instance of RSACryptoManager
     * @param client The parent client.
     * @param priv The private key of the client.
     * @param pubs An array of public keys for all clients
     */
    public RSACryptoManager(Client client, PrivateKey priv, PublicKey[] pubs) {
        this.client = client;
        this.priv = priv;
        this.pubs = pubs;
    }
    
    /** <P>Encrypts the secret share and the commitment value from the supplied share with the specified public key using RSA encryption.</P>
     * @param share The share with the secret share and commitment values that are to be encrypted.
     * @param shareNumber The share number of the client whose public key should be used in the encryption.
     * @throws CryptoException An indication of an error encrypting the information.
     * @return a <CODE>byte</CODE> array containing the encrypted version of the secret share and commitment value.
     */
    public byte[] encrypt(Share share, int shareNumber) throws CryptoException {
        BigInteger secret = share.getSecretShare();
        BigInteger commit = share.getSecretCommitment();
        
        //get an instance of the RSA cipher
        Cipher encryptCipher = null;
        int inputSize = 0;
        try {
            encryptCipher = Cipher.getInstance("RSA", "Cryptix");
            encryptCipher.initEncrypt(this.pubs[shareNumber - 1]);
            inputSize = encryptCipher.getInputBlockSize();
        } catch (NoSuchAlgorithmException e) {
            throw new CryptoException(e.toString());
        } catch (KeyException e) {
            throw new CryptoException(e.toString());
        } catch (NoSuchProviderException e) {
            throw new CryptoException(e.toString());
        }
        
        ByteArrayOutputStream bStream = new ByteArrayOutputStream();
        ObjectOutputStream out = null;
        
        try {
            out = new ObjectOutputStream(bStream);

            out.writeObject(secret);
            out.writeObject(commit);
            out.flush();
        } catch (IOException e) {
            throw new CryptoException("Error serializing values\n" + e.toString());
        }
        
        byte[] plainTextData = bStream.toByteArray();
        bStream = new ByteArrayOutputStream();
        
        try {
            out = new ObjectOutputStream(bStream);
            out.writeInt(plainTextData.length);
            out.close();
        } catch (IOException e) {
            throw new CryptoException("Error serializing data length\n" + e.toString());
        }
        
        byte[] plainTextDataLength = bStream.toByteArray();
        
        int paddingSize = inputSize - ((plainTextData.length + plainTextDataLength.length + 1) % inputSize);
        
        SecureRandom sr = new SecureRandom();
        byte[] padding = new byte[paddingSize];
        sr.nextBytes(padding); //gets the padding bytes
        
        byte[] plainText = new byte[plainTextData.length + plainTextDataLength.length + 1 + paddingSize];
        //length of the array is assumed to be less than 128
        //that corresponds to an (encoded) integer that is less than 128 bytes.
        plainText[0] = new Integer(plainTextDataLength.length).byteValue();
        int i;
        int j;
        int n;
        for(i=0;i < plainTextDataLength.length;i++) {
            plainText[i + 1] = plainTextDataLength[i];
        }
        for(i=0;i < plainTextData.length;i++) {
            plainText[i + plainTextDataLength.length + 1] = plainTextData[i];
        }
        for(i=0;i < paddingSize;i++) {
            plainText[i + plainTextDataLength.length + plainTextData.length + 1] = padding[i];
        }

        byte[] cipherText = null;
        try {
            cipherText = encryptCipher.crypt(plainText);
        } catch (Exception e) {
            throw new CryptoException(e.toString());
        }
        
        return cipherText;
    }
    
    /** <P>Decrypts the ciphertext into two BigIntegers using the local private key and places them into an array.</P>
     * @param cipherText The ciphertext created by {@link RSACryptoManager#encrypt(Share, int) encrypt()}.
     * @throws MalformedCryptoException An indication of invalid ciphertext.
     * @throws CryptoException An indication of error during the decryption process.
     * @return The secret share and commitment value as a <CODE>BigInteger</CODE> array.
     */    
    protected BigInteger[] decrypt(byte[] cipherText) throws MalformedCryptoException, CryptoException {
        Cipher decryptCipher = null;
        int inputSize = 0;
        try {
            decryptCipher = Cipher.getInstance("RSA", "Cryptix");
            decryptCipher.initDecrypt(this.priv);
            inputSize = decryptCipher.getInputBlockSize();
        } catch (Exception e) {
            throw new CryptoException("Error initializing signed data\n" + e.toString());
        }
        
        int i;
        byte[] plainText = null;
        try {
            plainText = decryptCipher.crypt(cipherText);
        } catch (Exception e) {
            throw new CryptoException(e.toString());
        }
        
        int intLength = new Byte(plainText[0]).intValue();
        byte[] dataLength = new byte[intLength];
        for(i = 0;i < intLength;i++) {
            dataLength[i] = plainText[i+1];
        }
        ByteArrayInputStream bIn = new ByteArrayInputStream(dataLength);
        ObjectInputStream in = null;
        int length;
        
        try {
            in = new ObjectInputStream(bIn);
            
            length = in.readInt();
            in.close();
        } catch (IOException e) {
            throw new MalformedCryptoException("Unable to read the data length\n" + e.toString());
        }
        
        byte[] data = new byte[length];
        for(i = 0;i < length;i++) {
            data[i] = plainText[i + intLength + 1];
        }
        
        BigInteger[] secretData = new BigInteger[2];
        bIn = new ByteArrayInputStream(data);
        try {
            in = new ObjectInputStream(bIn);
            
            secretData[0] = (BigInteger) in.readObject();
            secretData[1] = (BigInteger) in.readObject();
            in.close();
        } catch (Exception e) {
            throw new MalformedCryptoException("Error recovering secrets\n" + e.toString());
        }
        
        return secretData;
    }
    
    /** <P>Signs the data usa SHA-1 and RSA following PKCS#1 using the local private key.</P>
     * @param data The message data in a <CODE>byte</CODE> array.
     * @throws CryptoException An indication of an error during the signing process.
     * @return A <CODE>byte</CODE> array containing the serialized data and the digital signature.
     */    
    protected byte[] sign(byte[] data) throws CryptoException {
//        System.out.println("Signing - Start");
        Signature sigAlg = null;
        byte[] sig = null;
        int inputSize = 0;
        try {
            sigAlg = Signature.getInstance("SHA-1/RSA/PKCS#1","Cryptix");
            sigAlg.initSign(this.priv);
            sigAlg.update(data);
            sig = sigAlg.sign();
        } catch (Exception e) {
            throw new CryptoException("Error initializing signed data\n" + e.toString());
        }
        
        LinkedList message = new LinkedList();
        message.addLast(data);
        message.addLast(sig);
        
        ByteArrayOutputStream bOut = new ByteArrayOutputStream();
        try {
            ObjectOutputStream out = null;
            out = new ObjectOutputStream(bOut);

            out.writeObject(message);
            out.close();
        } catch (IOException e) {
            throw new CryptoException("Error creating signed message\n" + e.toString());
        }
        
//        System.out.println("Signing - end");
        
        return bOut.toByteArray();
    }
    
    /** <P>Verifies the signature with an attached message.</P>
     * @param messageBytes The message.
     * @param senderShareNumber The share number of the sender
     * @throws CryptoException An indication of an error during the verification process.
     * @throws MalformedCryptoException An indication that the message has an invalid signature.
     * @return The message bytes.
     */    
    protected byte[] validate(byte[] messageBytes, int senderShareNumber) throws CryptoException, MalformedCryptoException {
//        System.out.println("validate - start");
        if(messageBytes == null) {
            throw new MalformedCryptoException("messageBytes is null");
        }
        ByteArrayInputStream bIn = new ByteArrayInputStream(messageBytes);
        ObjectInputStream in = null;
        LinkedList message = null;
        
        try {
            in = new ObjectInputStream(bIn);
            
            message = (LinkedList) in.readObject();
            in.close();
        } catch (Exception e) {
            throw new CryptoException("Error reading message\n" + e.toString());
        }
        byte[] data = (byte[]) message.removeFirst();
        byte[] sig = (byte[]) message.removeFirst();
        
        Signature sigAlg = null;
        boolean isVerified = false;
        
        try {
            sigAlg = Signature.getInstance("SHA-1/RSA/PKCS#1","Cryptix");
            sigAlg.initVerify(this.pubs[senderShareNumber - 1]);
            sigAlg.update(data);
            isVerified = sigAlg.verify(sig);
        } catch (Exception e) {
            throw new CryptoException("Error creating signature object\n" + e.toString());
        }
        
        if(!isVerified) {
            throw new MalformedCryptoException("Signature invalid");
        }
        
//        System.out.println("validate - end");
        
        return data;
    }
    
    /** <P>Generates an RSA key pair.</P>
     * @return An RSA public/private key pair.
     */    
    public static KeyPair generateKeyPair() {
        BaseRSAKeyPairGenerator kpg = new BaseRSAKeyPairGenerator();
        kpg.initialize();
        return kpg.generateKeyPair();
    }
    
}
